import traceback
from typing import NamedTuple
from selenium import webdriver
from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as ec
from urllib.parse import urlparse
from time import sleep
from .strategy import Strategy
from ..chrome_cdp import CDP
from ..query import Query
from ..utils.logger import debug, info, warn, error
from ..events import Events, EventData


class Selectors:
    switch_selectors = False

    @property
    def container(self):
        return '.results__container.results__container--two-pane' if not Selectors.switch_selectors else \
            '.two-pane-serp-page__results-list'

    @property
    def jobs(self):
        return '.jobs-search__results-list li' if not Selectors.switch_selectors else \
            '.jobs-search__results-list li'

    @property
    def links(self):
        return '.jobs-search__results-list li a.result-card__full-card-link' if not Selectors.switch_selectors else \
            'a.base-card__full-link'

    @property
    def applyLink(self):
        return 'a[data-is-offsite-apply=true]'

    @property
    def dates(self):
        return 'time'

    @property
    def companies(self):
        return '.result-card__subtitle.job-result-card__subtitle' if not Selectors.switch_selectors else \
            '.base-search-card__subtitle'

    @property
    def places(self):
        return '.job-result-card__location' if not Selectors.switch_selectors else \
            '.job-search-card__location'

    @property
    def detailsPanel(self):
        return '.details-pane__content'

    @property
    def description(self):
        return '.description__text'

    @property
    def seeMoreJobs(self):
        return 'button.infinite-scroller__show-more-button'


class AnonymousStrategy(Strategy):
    def __init__(self, scraper: 'LinkedinScraper'):
        warn('AnonymousStrategy is no longer maintained and it won\'t probably work. It is recommended to use an authenticated session, see documentation at https://github.com/spinlud/py-linkedin-jobs-scraper#anonymous-vs-authenticated-session.')
        super().__init__(scraper)

    @staticmethod
    def __require_authentication(driver: webdriver) -> bool:
        """
        Verify if driver has been redirected to auth wall and needs authentication
        :param driver: webdriver
        :return: bool
        """

        parsed = urlparse(driver.current_url)
        return 'authwall' in parsed.path.lower()

    @staticmethod
    def __load_job_details(driver: webdriver, selectors: Selectors, job_id: str, timeout=2) -> object:
        """
        Wait for job details to load
        :param driver: webdriver
        :param job_id: str
        :param timeout: int
        :return: object
        """
        elapsed = 0
        sleep_time = 0.05

        while elapsed < timeout:
            loaded = driver.execute_script(
                '''
                    const detailsPanel = document.querySelector(arguments[1]);
                    const description = document.querySelector(arguments[2]);
                    return detailsPanel && detailsPanel.innerHTML.includes(arguments[0]) &&
                        description && description.innerText.length > 0;    
                ''',
                job_id,
                selectors.detailsPanel,
                selectors.description)

            if loaded:
                return {'success': True}

            sleep(sleep_time)
            elapsed += sleep_time

        return {'success': False, 'error': 'Timeout on loading job details'}

    @staticmethod
    def __load_more_jobs(driver: webdriver, selectors: Selectors, job_links_tot: int, timeout=2) -> object:
        """

        :param driver:
        :param job_links_tot:
        :param timeout:
        :return:
        """

        elapsed = 0
        sleep_time = 0.05
        clicked = False

        while elapsed < timeout:
            if not clicked:
                clicked = driver.execute_script(
                    '''
                        const button = document.querySelector(arguments[0]);
    
                        if (button) {
                            button.click();
                            return true;
                        }
                        else {
                            return false;
                        }    
                    ''',
                    selectors.seeMoreJobs)

            loaded = driver.execute_script(
                '''
                    window.scrollTo(0, document.body.scrollHeight);
                    return document.querySelectorAll(arguments[0]).length > arguments[1];
                ''',
                selectors.jobs,
                job_links_tot)

            if loaded:
                return {'success': True}

            sleep(sleep_time)
            elapsed += sleep_time

        return {'success': False, 'error': 'Timeout on loading more jobs'}

    @staticmethod
    def __accept_cookies(driver: webdriver, tag: str) -> None:
        """
        Accept cookies
        :param driver:
        :param tag:
        :return:
        """

        try:
            driver.execute_script(
                '''
                    const buttons = Array.from(document.querySelectorAll('button'));
                    const cookieButton = buttons.find(e => e.innerText.includes('Accept cookies'));
                    
                    if (cookieButton) {
                        cookieButton.click();
                    }
                '''
            )
        except:
            debug(tag, 'Failed to accept cookies')

    def run(
        self,
        driver: webdriver,
        cdp: CDP,
        search_url: str,
        query: Query,
        location: str,
        page_offset: int,
    ) -> None:
        """
        Run scraper
        :param driver: webdriver
        :param cdp: CDP
        :param search_url: str
        :param query: Query
        :param location: str
        :param page_offset: int
        :return: None
        """

        tag = f'[{query.query}][{location}]'
        processed = 0

        info(tag, f'Opening {search_url}')
        driver.get(search_url)

        # Verify if redirected to auth wall
        if AnonymousStrategy.__require_authentication(driver):
            error('Scraper failed to run in anonymous mode, authentication may be necessary for this environment. '
                  'Please check the documentation on how to use an authenticated session.')
            return

        # Linkedin seems to randomly load two different set of selectors:
        # the following hack tries to switch between the two sets
        Selectors.switch_selectors = False
        selectors = Selectors()

        try:
            info(tag, 'Trying first selectors set')
            debug(tag, 'Waiting selector', selectors.container)
            WebDriverWait(driver, 3).until(ec.presence_of_element_located((By.CSS_SELECTOR, selectors.container)))
            # First set of selectors confirmed
        except:
            try:
                # Try to load second set of selectors
                info(tag, 'Trying second selectors set')
                Selectors.switch_selectors = True
                debug(tag, 'Waiting selector', selectors.container)
                WebDriverWait(driver, 3).until(ec.presence_of_element_located((By.CSS_SELECTOR, selectors.container)))
                # Second set of selectors confirmed
            except:
                info(tag, 'Failed to load container selector, skip')
                return

        job_index = 0

        info(tag, 'OK')
        info(tag, 'Starting pagination loop')

        # Pagination loop
        while processed < query.options.limit:
            AnonymousStrategy.__accept_cookies(driver, tag)

            jobs_tot = driver.execute_script('return document.querySelectorAll(arguments[0]).length;', selectors.jobs)

            if jobs_tot == 0:
                info(tag, 'No jobs found, skip')
                break

            info(tag, f'Found {jobs_tot} jobs')

            # Jobs loop
            while job_index < jobs_tot and processed < query.options.limit:
                sleep(self.scraper.slow_mo)
                tag = f'[{query.query}][{location}][{processed + 1}]'

                # Extract job main fields and navigate job link
                debug(tag, 'Evaluating selectors', [
                    selectors.jobs,
                    selectors.links,
                    selectors.companies,
                    selectors.places,
                    selectors.dates])

                try:
                    job_id, job_link, job_title, job_company, job_place, job_date = driver.execute_script(
                        '''
                            const index = arguments[0];
                            const job = document.querySelectorAll(arguments[1])[index];
                            const link = job.querySelector(arguments[2]);
                            
                            // Click job link and scroll
                            link.scrollIntoView();
                            link.click();
                            const linkUrl = link.getAttribute("href");
                            
                            // Extract job id
                            let jobId = '';
                            
                            // First set of selectors                            
                            jobId = job.getAttribute('data-id');
                            
                            // Second set of selectors
                            if (!jobId) {
                                jobId = job.querySelector(arguments[2])
                                    .parentElement.getAttribute('data-entity-urn').split(':').splice(-1)[0];
                            }
                                                                                
                            return [
                                jobId,
                                linkUrl,
                                job.querySelector(arguments[2]).innerText,
                                job.querySelector(arguments[3]).innerText,
                                job.querySelector(arguments[4]).innerText,
                                job.querySelector(arguments[5]).getAttribute('datetime')
                            ];
                        ''',
                        job_index,
                        selectors.jobs,
                        selectors.links,
                        selectors.companies,
                        selectors.places,
                        selectors.dates)

                    # Wait for job details to load
                    debug(tag, f'Loading details of job {job_id}')
                    load_result = AnonymousStrategy.__load_job_details(driver, selectors, job_id)

                    if not load_result['success']:
                        error(tag, load_result['error'])
                        job_index += 1
                        continue

                    # Extract description
                    debug(tag, 'Evaluating selectors', [selectors.description])

                    job_description, job_description_html = driver.execute_script(
                        '''
                            const el = document.querySelector(arguments[0]);
                        
                            return [
                                el.innerText,
                                el.outerHTML    
                            ];
                        ''',
                        selectors.description)

                    # Extract apply link
                    debug(tag, 'Evaluating selectors', [selectors.applyLink])

                    job_apply_link = driver.execute_script(
                        '''
                            const applyBtn = document.querySelector(arguments[0]);
                            return applyBtn ? applyBtn.getAttribute("href") : '';
                        ''',
                        selectors.applyLink)

                except BaseException as e:
                    error(tag, e, traceback.format_exc())
                    self.scraper.emit(Events.ERROR, str(e) + '\n' + traceback.format_exc())
                    job_index += 1
                    continue

                data = EventData(
                    query=query.query,
                    location=location,
                    job_id=job_id,
                    job_index=job_index,
                    title=job_title,
                    company=job_company,
                    place=job_place,
                    date=job_date,
                    link=job_link,
                    apply_link=job_apply_link,
                    description=job_description,
                    description_html=job_description_html)

                info(tag, 'Processed')

                job_index += 1
                processed += 1

                self.scraper.emit(Events.DATA, data)

                # Try fetching more jobs
                if processed < query.options.limit and job_index == jobs_tot:
                    jobs_tot = driver.execute_script('return document.querySelectorAll(arguments[0]).length;',
                                                     selectors.jobs)

            # Check if we reached the limit of jobs to process
            if processed == query.options.limit:
                break

            # Check if we need to paginate
            info(tag, 'Checking for new jobs to load...')
            load_result = AnonymousStrategy.__load_more_jobs(driver, selectors, jobs_tot)

            if not load_result['success']:
                info(tag, "Couldn't find more jobs for the running query")
                break



import threading
import json
import websocket
import ssl
import time
from time import sleep
from websocket import WebSocketTimeoutException, WebSocketConnectionClosedException
from types import FunctionType
from typing import Callable, Union, List
from .request import CDPRequest
from .response import CDPResponse
from .target_info import CDPTargetsInfoResponse
from .events import Events
from .cookie import CDPCookie
from ..utils.logger import debug, info, warn, error


class CDP:
    def __init__(self, ws_url, timeout=1):
        self.ws_url = ws_url
        self.timeout = timeout
        self._tag = f'[T_{threading.get_ident()} ChromeDevTools]'
        self._stop = threading.Event()
        self._ws = None
        self._ws_loop_th = None
        self._is_running = False
        self._id = 0
        self.targets_info_response = None
        # self._responses = {}

        def __default_request_handler(request: CDPRequest) -> None:
            request.resume()

        def __default_response_handler(response: CDPResponse) -> None:
            pass

        self._event_handlers = {
            'request': __default_request_handler,
            'response': __default_response_handler,
        }

    def __ws_loop(self):
        """
        Loop to receive messages from websocket server
        :return:
        """

        while not self._stop.is_set():
            try:
                msg = self._ws.recv()
                parsed = json.loads(msg)

                if 'error' in parsed:
                    print(self._tag, '[ERROR]', parsed)

                # print(parsed)

                # Intercept request/response
                if 'method' in parsed:
                    event = parsed['method']

                    # Request handler
                    if event == Events.REQUEST.value:
                        if self._event_handlers['request'] is not None:
                            request = CDPRequest(self, parsed)
                            self._event_handlers['request'](request)

                    # Response handler
                    if event == Events.RESPONSE.value:
                        if self._event_handlers['response'] is not None:
                            response = CDPResponse(self, parsed)
                            self._event_handlers['response'](response)

                # Intercept Target.getTargets
                if 'result' in parsed and 'targetInfos' in parsed['result']:
                    target_info_response = CDPTargetsInfoResponse(self, parsed)
                    self.targets_info_response = target_info_response
            except (WebSocketTimeoutException, WebSocketConnectionClosedException) as e:
                continue

    def call_method(self, method: str, **params) -> int:
        """
        Call dev tools method with the given parameters
        :param method: str
        :param params:
        :return: Id of the request
        """

        if not self._ws or not self._ws.connected:
            raise RuntimeError(self._tag + '\tWebsocket not connected')

        self._id += 1

        msg = {'id': self._id, 'method': method, 'params': params}
        debug(self._tag, 'Calling method', msg)
        self._ws.send(json.dumps(msg))
        return self._id

    def start(self) -> None:
        """
        Start ChromeDevTools client
        :return: None
        """

        if self._is_running:
            raise RuntimeError(self._tag, 'It is already running')

        debug(self._tag, 'Connecting to websocket', self.ws_url)

        self._ws = websocket.create_connection(
            self.ws_url,
            enable_multithread=True,
            skip_utf8_validation=True,
            sslopt={
                'check_hostname': False,
                'cert_reqs': ssl.CERT_NONE
            },
        )

        self._ws.settimeout(self.timeout)

        # Enable Fetch domain
        self.call_method('Fetch.enable')

        # Enable Network domain
        self.call_method('Network.enable')

        self._stop.clear()
        self._ws_loop_th = threading.Thread(target=self.__ws_loop, daemon=True)
        debug(self._tag, 'Starting websocket loop thread', self._ws_loop_th.ident)
        self._ws_loop_th.start()
        self._is_running = True

    def stop(self) -> None:
        """
        Stop ChromeDevTools client
        :return: None
        """

        self._stop.set()

        if self._ws_loop_th:
            self._ws_loop_th.join()

        self._is_running = False

        debug(self._tag, 'Closing websocket')
        self._ws.close()

    def on(self, event: str, cb: Union[Callable, None]) -> None:
        """
        Override event handler
        :param event: str
        :param cb: Callable | None
        :return: None
        """

        if cb is not None and not isinstance(cb, FunctionType):
            raise ValueError(self._tag + '\tCallback must be a function')

        if event not in self._event_handlers.keys():
            raise ValueError(self._tag + f'\tEvent must be one of ({", ".join(self._event_handlers.keys())})')

        self._event_handlers[event] = cb

    def set_user_agent(self, ua: str) -> None:
        """
        Set user agent
        :param ua: str
        :return: None
        """

        debug(self._tag, 'Setting user agent', ua)
        self.call_method('Network.setUserAgentOverride', userAgent=ua)

    def set_bypass_csp(self, v: bool):
        """
        Enable page Content Security Policy by-passing.
        :param v: bool
        :return: None
        """

        debug(self._tag, 'Setting Content Security Policy by-passing', v)
        self.call_method('Page.setBypassCSP', enabled=v)

    def get_cookies(self):
        """
        Returns all browser cookies for the current URL
        :return:
        """

        debug(self._tag, 'Getting browser cookies')
        req_id = self.call_method('Network.getCookies')
        # TODO how to get cookies from websocket?

    def set_cookies(self, cookies: List[CDPCookie]):
        """
        Returns all browser cookies for the current URL
        :return:
        """

        debug(self._tag, 'Setting cookies', cookies)
        self.call_method('Network.setCookies', cookies=[e.to_dict() for e in cookies])

    def set_discover_targets(self):
        """
        Controls whether to discover available targets and notify via targetCreated/targetInfoChanged/targetDestroyed events.
        :return:
        """

        debug(self._tag, 'Target.setDiscoverTargets')
        self.call_method('Target.setDiscoverTargets')

    def get_targets(self, timeout=2) -> object:
        """
        Retrieves a list of available targets.
        :return:
        """

        debug(self._tag, 'Getting targets')
        req_id = self.call_method('Target.getTargets')

        sleep_time = 0.05
        elapsed = 0

        while elapsed < timeout:
            if self.targets_info_response and self.targets_info_response.id == req_id:
                return {'success': True, 'result': self.targets_info_response}

            sleep(sleep_time)
            elapsed += sleep_time

        return {'success': False, 'error': 'Timeout'}

    def close_target(self, target_id) -> None:
        """
        Closes the target. If the target is a page that gets closed too.
        :return:
        """

        self.call_method('Target.closeTarget', targetId=target_id)




from selenium import webdriver
from ..chrome_cdp import CDP
from ..query import Query


class Strategy:
    def __init__(self, scraper: 'LinkedinScraper'):
        self.scraper = scraper

    def run(
        self,
        driver: webdriver,
        cdp: CDP,
        search_url: str,
        query: Query,
        location: str,
        page_offset: int
    ) -> None:
        raise NotImplementedError('Must implement method in subclass')



class CDPTargetsInfoResponse:
    def __init__(self, cdp, message):
        self._cdp = cdp
        self._tag = '[CDPTargetsInfoResponse]'

        self.id = message.pop('id', None)
        self.targets = [CDPTargetInfo(cdp, e) for e in message['result']['targetInfos']]

    def __str__(self):
        return f'{self._tag} id={self.id} targets={[str(e) for e in self.targets]}'


class CDPTargetInfo:
    def __init__(self, cdp, message):
        self._tag = '[CDPTargetInfo]'
        self.targetId = message.pop('targetId', None)
        self.type = message.pop('type', None)
        self.title = message.pop('title', None)
        self.url = message.pop('url', None)
        self.attached = message.pop('attached', None)
        self.openerId = message.pop('openerId', None)
        self.canAccessOpener = message.pop('canAccessOpener', None)
        self.openerFrameId = message.pop('openerFrameId', None)
        self.browserContextId = message.pop('browserContextId', None)

    def __str__(self):
        return f'{self._tag} targetId={self.targetId} type={self.type} title={self.title} url={self.url} attached={self.attached}'



import json
from urllib.request import urlopen
from selenium import webdriver
from selenium.webdriver.common.proxy import Proxy, ProxyType
from selenium.webdriver.chrome.options import Options
from linkedin_jobs_scraper.utils.logger import debug, info


def get_default_driver_options(width=1472, height=828, headless=True) -> Options:
    """
    Generate default Chrome driver options
    :param width: int
    :param height: int
    :param headless: bool
    :return: Options
    """

    chrome_options = Options()
    chrome_options.headless = headless
    chrome_options.page_load_strategy = 'none'

    chrome_options.add_argument("--enable-automation"),
    chrome_options.add_argument("--start-maximized"),
    chrome_options.add_argument(f"--window-size={width},{height}"),
    chrome_options.add_argument("--lang=en-GB"),
    chrome_options.add_argument("--no-sandbox"),
    chrome_options.add_argument("--disable-setuid-sandbox"),
    chrome_options.add_argument("--disable-dev-shm-usage"),
    chrome_options.add_argument("--disable-gpu"),
    chrome_options.add_argument("--disable-accelerated-2d-canvas"),
    # chrome_options.add_argument("--proxy-server='direct://"),
    # chrome_options.add_argument("--proxy-bypass-list=*"),
    chrome_options.add_argument("--allow-running-insecure-content"),
    chrome_options.add_argument("--disable-web-security"),
    chrome_options.add_argument("--disable-client-side-phishing-detection"),
    chrome_options.add_argument("--disable-notifications"),
    chrome_options.add_argument("--mute-audio"),
    chrome_options.add_argument("--ignore-certificate-errors"),
    chrome_options.add_argument("--remote-allow-origins=*"),

    # Disable downloads
    chrome_options.add_experimental_option(
        'prefs', {
            'safebrowsing.enabled': 'false',
            'download.prompt_for_download': False,
            'download.default_directory': '/dev/null',
            'download_restrictions': 3,
            'profile.default_content_setting_values.notifications': 2,
        }
    )

    return chrome_options


def get_driver_proxy_capabilities(proxy: str):
    """
    Use a single proxy directly from the browser
    :param proxy:
    :return:
    """

    proxy = Proxy()
    proxy.proxy_type = ProxyType.MANUAL
    proxy.http_proxy = proxy
    proxy.ssl_proxy = proxy
    proxy.ftp_proxy = proxy
    proxy.auto_detect = False
    capabilities = webdriver.DesiredCapabilities.CHROME.copy()
    proxy.add_to_capabilities(capabilities)
    return capabilities


def build_driver(executable_path: str = None, options: Options = None, headless=True, timeout=20) -> webdriver:
    """
    Build Chrome driver instance
    :param executable_path: str
    :param options: Options
    :param headless: bool
    :param timeout: int
    :return: webdriver
    """

    kwargs = {}

    if executable_path is not None:
        kwargs['executable_path'] = executable_path

    kwargs['options'] = options if options is not None else get_default_driver_options(headless=headless)
    # kwargs['desired_capabilities'] = get_driver_proxy_capabilities('http://localhost:8888')

    driver = webdriver.Chrome(**kwargs)
    driver.set_page_load_timeout(timeout)

    return driver


def get_debugger_url(driver: webdriver) -> str:
    """
    Get Chrome debugger url
    :param driver: webdriver
    :return: str
    """

    chrome_debugger_url = f"http://{driver.capabilities['goog:chromeOptions']['debuggerAddress']}"
    debug('Chrome Debugger Url', chrome_debugger_url)
    return chrome_debugger_url


def get_websocket_debugger_url(driver: webdriver) -> str:
    """
    Get Chrome websocket debugger url
    :param driver: webdriver
    :return: str
    """

    chrome_debugger_url = get_debugger_url(driver)
    info('Chrome debugger url', chrome_debugger_url)
    response = json.loads(urlopen(f'{chrome_debugger_url}/json').read().decode())
    debug('Websocket debugger url', response[0]['webSocketDebuggerUrl'])
    return response[0]['webSocketDebuggerUrl']



from urllib.parse import urlparse, urlencode, parse_qsl, urljoin


def get_query_params(url: str) -> dict:
    """
    Extract url query parameters as a dictionary
    :param url:
    :return: dict
    """

    parsed = urlparse(url)
    return dict(parse_qsl(parsed.query))


def get_url_no_query_params(url: str) -> str:
    """
    Returns url without query parameters
    :param url:
    :return:
    """

    parsed = urlparse(url)
    parsed = parsed._replace(query='')
    return parsed.geturl()


def override_query_params(url: str, override_params: dict) -> str:
    """
    Override url query parameters
    :param url:
    :param override_params:
    :return:
    """

    params = get_query_params(url)

    for k, v in override_params.items():
        params[k] = v

    return urlparse(url)._replace(query=urlencode(params)).geturl()


def get_domain(url: str) -> str:
    """
    Return SLD (Second Level Domain) from url
    :param url: str
    :return: str
    """

    return '.'.join(urlparse(url).netloc.split('.')[-2:])


def get_location(url: str) -> str:
    """
    Return location from url (with scheme)
    :param url: str
    :return: str
    """

    parsed = urlparse(url)
    return f'{parsed.scheme}://{parsed.netloc}'
